from rest_framework import viewsets
from rest_framework import serializers

import rest_framework_extras

from link.models import Link


class PropertiesMixin(serializers.Serializer):
    absolute_url = serializers.ReadOnlyField()

    class Meta(object):
        fields = ("absolute_url", )


class LinkSerializer(PropertiesMixin, serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = Link
        fields = ("absolute_url", )


class LinkObjectsViewSet(viewsets.ModelViewSet):
    queryset = Link.objects.all()
    serializer_class = LinkSerializer


def register(router):
    return rest_framework_extras.register(
        router, (("link-link", LinkObjectsViewSet),)
    )



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
try:
    from django.urls import reverse, NoReverseMatch
except ImportError:
    from django.core.urlresolvers import reverse, NoReverseMatch

class ViewParam(models.Model):
    key = models.CharField(max_length=64)
    value = models.CharField(max_length=256)

    class Meta:
        ordering = ["key"]

    def __unicode__(self):
        return "%s:%s" % (self.key, self.value)

    __str__ = __unicode__


class Link(models.Model):
    title = models.CharField(
        max_length=256,
        help_text="A short descriptive title."
    )
    slug = models.SlugField(
        max_length=256,
        db_index=True,
    )
    view_name = models.CharField(
        max_length=256, blank=True, null=True,
        help_text="View name to which this link will redirect."
    )
    view_params = models.ManyToManyField(
        ViewParam, blank=True, null=True
    )
    target_content_type = models.ForeignKey(
        ContentType, blank=True, null=True,
        related_name="link_target_content_type",
        on_delete=models.CASCADE
    )
    target_object_id = models.PositiveIntegerField(blank=True, null=True)
    target = GenericForeignKey("target_content_type", "target_object_id")
    url = models.CharField(
        max_length=256, blank=True, null=True,
        help_text="URL to which this link will redirect."
    )

    class Meta:
        ordering = ["title"]

    def __unicode__(self):
        return self.title

    __str__ = __unicode__

    @property
    def absolute_url(self):
        return self.get_absolute_url()

    def get_absolute_url(self):
        """
        Returns URL to which link should redirect based on a reversed view
        name, category, target or explicitly provided URL.
        """
        if self.view_name:
            kwargs = dict(
                (param.key, param.value) for param in self.view_params.all()
            )
            try:
                return reverse(self.view_name, kwargs=kwargs)
            except NoReverseMatch:
                pass
        elif self.target:
            try:
                return self.target.get_absolute_url()
            except AttributeError:
                pass
        else:
            # Django can be served in a subdirectory. Transparently fix urls.
            if "://" in self.url:
                return self.url

            # Urls not starting with a slash probably do so with reason. Skip.
            if not self.url.startswith("/"):
                return self.url

            # Request is not available here so use reverse to determine root
            try:
                root = reverse("home")
            except NoReverseMatch:
                return self.url

            # /abc and /root/abc must be transformed into /root/abc
            if not self.url.startswith(root):
                return root.rstrip("/") + "/" + self.url.lstrip("/")

        return self.url



from django import template

from link.models import Link

register = template.Library()


@register.inclusion_tag(
    "link/inclusion_tags/link_detail.html", takes_context=True
)
def render_link(context, slug):
    try:
        context["object"] = Link.objects.get(slug=slug)
    except Link.DoesNotExist:
        pass
    return context



from django import forms


from link.models import Link
from link.utils import get_view_name_choices


class LinkAdminForm(forms.ModelForm):
    view_name = forms.ChoiceField(required=False)

    class Meta:
        model = Link
        fields = [
            "title", "slug", "view_name", "view_params", "target_content_type",
            "target_object_id", "url"
        ]

    def __init__(self, *args, **kwargs):
        """
        Set view_name choices to the availbale url pattern names.
        """
        super(LinkAdminForm, self).__init__(*args, **kwargs)

        view_name_choices = [("", "-- Select --"), ] + get_view_name_choices()
        self.fields["view_name"].choices = view_name_choices

    def clean(self):
        """
        Adds validation to ensure that only one view_name, target or url is
        selected.
        """
        cleaned_data = super(LinkAdminForm, self).clean()
        field_count = 0
        for fieldname in ["view_name", "target_content_type", "url"]:
            if cleaned_data.get(fieldname):
                if field_count:
                    raise forms.ValidationError(
                        "You may set at most one of view_name, target or URL."
                    )
                field_count += 1

        if not field_count:
            raise forms.ValidationError(
                "You may set at least one of view_name, target or URL."
            )

        return cleaned_data



from importlib import import_module

from django.conf import settings
try:
    from django.urls import URLResolver as RegexURLResolver
    from django.urls import URLPattern as RegexURLPattern
except ImportError:
    from django.core.urlresolvers import RegexURLResolver, RegexURLPattern

from link import SETTINGS


def get_view_names(urlpatterns=None, view_names=[], namespace=None):
    if urlpatterns is None:
        urlpatterns = import_module(settings.ROOT_URLCONF).urlpatterns
    if namespace not in SETTINGS.get("excluded-viewname-choices"):
        for pattern in urlpatterns:
            if isinstance(pattern, RegexURLResolver):
                get_view_names(
                    pattern.url_patterns, view_names, pattern.namespace
                )
            elif isinstance(pattern, RegexURLPattern):
                view_name = pattern.name
                if view_name:
                    if namespace:
                        view_name = "%s:%s" % (namespace, view_name)
                    if view_name not in view_names:
                        view_names.append(view_name)
    return view_names


def get_view_name_choices():
    return [tuple([view_name, view_name]) for view_name in get_view_names()]



from django.contrib import admin

from link.forms import LinkAdminForm
from link.models import Link, ViewParam


class LinkAdmin(admin.ModelAdmin):
    form = LinkAdminForm
    list_display = ("title", "_get_absolute_url")
    prepopulated_fields = {"slug": ["title"]}
    filter_horizontal = ["view_params"]
    search_fields = [
        "title", "slug", "url"
    ]

    def _get_absolute_url(self, obj):
        url = obj.get_absolute_url()
        if url:
            return "<a href=\"%s\" target=\"public\">%s</a>" % (url, url)
        return "<p class=\"errornote\">Inactive or broken link</p>"

    _get_absolute_url.short_description = "Permalink"
    _get_absolute_url.allow_tags = True


admin.site.register(Link, LinkAdmin)
admin.site.register(ViewParam)



from django.test import TestCase

from link import utils


class UtilsTestCase(TestCase):
    def test_get_view_names(self):
        self.assertEqual(
            utils.get_view_names(),
            ["link-1", "link-2", "link:link-list", "link:link-detail"]
        )



from django.test import TestCase

from link import forms


class FormTestCase(TestCase):
    def setUp(self):
        self.form_data = {
            "title": "Link 1 Title",
            "slug": "link-1-title"
        }

    def test_admin_form(self):

        # ensure that not selecting a target raises a validation error
        admin_form = forms.LinkAdminForm(self.form_data)
        admin_form.full_clean()
        self.assertEqual(len(admin_form.errors["__all__"]), 1)

        # ensure the form is valid
        self.form_data["url"] = "/test-link/"
        admin_form = forms.LinkAdminForm(self.form_data)
        self.assertTrue(admin_form.is_valid())
        self.assertEqual(len(admin_form.errors), 0)

        # ensure that selecting multiple targets fails
        self.form_data["view_name"] = "link-1"
        admin_form = forms.LinkAdminForm(self.form_data)
        admin_form.full_clean()
        self.assertEqual(len(admin_form.errors["__all__"]), 1)

    def tearDown(self):
        pass



from django.test import TestCase
from django.test.client import Client

from link import models


class ViewTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.link_data = {
            "title": "Link 1 Title",
            "slug": "link-1-title",
            "url": "/link-1/"
        }
        self.link = models.Link.objects.create(**self.link_data)

    def test_detail(self):
        response = self.client.get("/link/%s/" % self.link_data["slug"])
        self.assertContains(response, self.link_data["title"])

    def test_list(self):
        link_data2 = {
            "title": "Link 2 Title",
            "slug": "link-2-title",
            "url": "/link-2/"
        }
        models.Link.objects.create(**link_data2)

        response = self.client.get("/link/")
        self.assertContains(response, self.link_data["title"])
        self.assertContains(response, link_data2["title"])

    def tearDown(self):
        pass



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-09-02 02:49
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('link', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ViewParam',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=64)),
                ('value', models.CharField(max_length=256)),
            ],
        ),
        migrations.AddField(
            model_name='link',
            name='view_params',
            field=models.ManyToManyField(blank=True, null=True, to='link.ViewParam'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-11-24 10:49
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('link', '0002_auto_20160902_0249'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='viewparam',
            options={'ordering': ['key']},
        ),
    ]



from django.contrib.auth import get_user_model
from django.test import TestCase
from django.test.client import Client


class AdminTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        super(AdminTestCase, cls).setUpTestData()
        cls.editor = get_user_model().objects.create(
            username="editor",
            email="editor@test.com",
            is_superuser=True,
            is_staff=True
        )
        cls.editor.set_password("password")
        cls.editor.save()

    def setUp(self):
        super(AdminTestCase, self).setUp()
        self.client.login(username="editor", password="password")

    def test_admin(self):
        response = self.client.get("/admin/")
        self.assertEqual(response.status_code, 200)

    def test_admin_link(self):

        # ensure the link add page can be reached
        response = self.client.get("/admin/link/link/add/")
        self.assertEqual(response.status_code, 200)

    def tearDown(self):
        self.client.logout()



from django.conf.urls import include, url
from django.contrib import admin

from link.tests import views


urlpatterns = [
    url(r"^admin/", admin.site.urls),
    url(r"^link/1/$", views.TestView.as_view(), name="link-1"),
    url(r"^link/2/$", views.TestView.as_view(), name="link-2"),
    url(r'^link/', include("link.urls", namespace="link"))
]



from django.views import generic

from link.models import Link


class LinkDetailView(generic.detail.DetailView):
    model = Link
    template = "link/link_detail.html"


class LinkListView(generic.list.ListView):
    model = Link
    template = "link/link_list.html"



from setuptools import setup, find_packages

description_files = ["README.rst", "AUTHORS.rst", "CHANGELOG.rst"]

setup(
    name="django-link",
    description="Modeled link objects which include linking to objects, views and static urls.",
    long_description="".join([open(f, "r").read() for f in description_files]),
    version="2.0.0",
    author="Praekelt Consulting",
    author_email="dev@praekelt.com",
    license="BSD",
    url="http://github.com/praekelt/django-link",
    packages=find_packages(),
    dependency_links=[],
    install_requires=[
        "django",
        "djangorestframework-extras>=0.3"
    ],
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    zip_safe=False,
    include_package_data=True
)



from django.test import TestCase

from link import models


class ModelTestCase(TestCase):
    def setUp(self):
        self.link_data = {
            "title": "Link 1 Title",
            "url": "/link-1/"
        }
        self.viewparam_data = {
            "key": "slug",
            "value": "1",
        }

    def test_link(self):
        link = models.Link.objects.create(**self.link_data)

        # ensure the data was saved correctly
        for key, value in self.link_data.items():
            self.assertEqual(getattr(link, key), value)

        # ensure get_absolute_url returns the correct string
        self.assertEqual(link.absolute_url, "/link-1/")

        # ensure that links with view names selected render the correct string
        link.view_name = "link-2"
        self.assertEqual(link.absolute_url, "/link/2/")

        # ensure that links with target objects render the correct string
        content_link_data = self.link_data.copy()
        content_link_data["url"] = "/content/1/"
        content_link = models.Link.objects.create(**content_link_data)
        link.view_name = None
        link.target = content_link
        self.assertEqual(link.absolute_url, "/content/1/")

    def test_viewparam(self):
        link = models.Link.objects.create(**self.link_data)
        viewparam = models.ViewParam.objects.create(**self.viewparam_data)
        link.view_params.add(viewparam)
        link.view_name = "link:link-detail"

        # ensure the url is returned correctly
        self.assertEqual(link.absolute_url, "/link/1/")



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-09-01 09:28
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Link',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(help_text=b'A short descriptive title.', max_length=256)),
                ('slug', models.SlugField(max_length=256)),
                ('view_name', models.CharField(blank=True, help_text=b'View name to which this link will redirect.', max_length=256, null=True)),
                ('target_object_id', models.PositiveIntegerField(blank=True, null=True)),
                ('url', models.CharField(blank=True, help_text=b'URL to which this link will redirect.', max_length=256, null=True)),
                ('target_content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='link_target_content_type', to='contenttypes.ContentType')),
            ],
            options={
                'ordering': ['title'],
            },
        ),
    ]



from django.conf.urls import url

from link.views import LinkDetailView, LinkListView


app_name = "link"
urlpatterns = [
    url(r'^$', LinkListView.as_view(), name="link-list"),
    url(r'^(?P<slug>[-\w]+)/$', LinkDetailView.as_view(), name="link-detail")
]
